;==============================================================================
;
; V20-MBC (HW ref. A250220) CP/M 2.2 (8080) Custom BIOS - S140520
;
;
; NOTE 1: This CBIOS requires IOS S260320-R180520 (or following revisions until 
;         otherwise stated).
;
; NOTE 2: Adapted from the Z80-MBC2 (HW ref. A040618) BIOS S030818-R140318.
;         I left the original Z80 code in the comments.
;
;
; CHANGELOG:
;
; S140520           First release.
;
;==============================================================================

; CP/M addresses
CCP         = CBASE          ; CP/M System entry
BDOS        = CCP + $806     ; BDOS entry
BIOS        = CCP + $1600    ; BIOS entry
IOBYT       = $0003          ; IOBYTE address
CDISK       = $0004          ; Address of Current drive name and user number
CCPLEN      = CBASE + 7      ; Address of current number of chars into the CCP input buffer
CCPFIRS     = CBASE + 8      ; Address of the first charater of the CCP input buffer

; BIOS equates
NDISKS      .equ    16       ; Number of Disk Drives

; Commons ASCII chars
eos         .equ    $00      ; End of string
cr          .equ    $0d      ; Carriage return
lf          .equ    $0a      ; Line feed

; V20-MBC IOS equates
EXC_WR_OPCD .equ    $00      ; Address of the EXECUTE WRITE OPCODE write port
EXC_RD_OPCD .equ    $00      ; Address of the EXECUTE READ OPCODE read port
STO_OPCD    .equ    $01      ; Address of the STORE OPCODE write port
SERIAL_RX   .equ    $01      ; Address of the SERIAL RX read port
SYSFLAGS    .equ    $02      ; Address of the SYSFLAGS read port
SERTX_OPC   .equ    $01      ; SERIAL TX opcode
SELDISK_OPC .equ    $09      ; SELDISK opcode
SELTRCK_OPC .equ    $0A      ; SELTRACK opcode
SELSECT_OPC .equ    $0B      ; SELSECT opcode
WRTSECT_OPC .equ    $0C      ; WRITESECT opcode
ERRDSK_OPC  .equ    $85      ; ERRDISK opcode
RDSECT_OPC  .equ    $86      ; READSECT opcode
SDMOUNT_OPC .equ    $87      ; SDMOUNT opcode

; Blocking and deblocking algorithm equates
; (see CP/M 2.2 Alteration Guide p.34 and APPENDIX G)
blksiz      .equ    4096            ;CP/M allocation size
hstsiz      .equ    512             ;host disk sector size
hstspt      .equ    32              ;host disk sectors/trk
hstblk      .equ    hstsiz/128      ;CP/M sects/host buff
cpmspt      .equ    hstblk * hstspt ;CP/M sectors/track
secmsk      .equ    hstblk-1        ;sector mask
wrall       .equ    0               ;write to allocated
wrdir       .equ    1               ;write to directory
wrual       .equ    2               ;write to unallocated


    .org BIOS

; =========================================================================== ;
;                                                                             ;
; BIOS jump table                                                             ;
;                                                                             ;
; =========================================================================== ;

BOOT
    jmp BOOT_    ; COLD START
WBOOT
WBOOTE
    jmp WBOOT_   ; WARM START
CONST
    jmp CONST_   ; CONSOLE STATUS
CONIN
    jmp CONIN_   ; CONSOLE CHARACTER IN
CONOUT
    jmp CONOUT_  ; CONSOLE CHARACTER OUT
LIST
    jmp LIST_    ; LIST CHARACTER OUT
PUNCH
    jmp PUNCH_   ; PUNCH CHARACTER OUT
READER
    jmp READER_  ; READER CHARACTER OUT
HOME
    jmp HOME_    ; MOVE HEAD TO HOME POSITION
SELDSK
    jmp SELDSK_  ; SELECT DISK
SETTRK
    jmp SETTRK_  ; SET TRACK NUMBER
SETSEC
    jmp SETSEC_  ; SET SECTOR NUMBER
SETDMA
    jmp SETDMA_  ; SET DMA ADDRESS
READ
    jmp READ_    ; READ DISK
WRITE
    jmp WRITE_   ; WRITE DISK
PRSTAT
    jmp LISTST_  ; RETURN LIST STATUS
SECTRN
    jmp SECTRN_  ; SECTOR TRANSLATE
    

; =========================================================================== ;
;                                                                             ;
; Disk parameter headers for disk 0 to 15                                     ;
;                                                                             ;
; =========================================================================== ;


dpbase
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb0,0000h,alv00
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv01
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv02
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv03
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv04
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv05
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv06
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv07
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv08
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv09
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv10
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv11
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv12
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv13
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv14
        .DW 0000h,0000h,0000h,0000h,dirbuf,dpb,0000h,alv15

; First drive has one reserved track for CP/M
dpb0
        .DW 128 ;SPT - 128 bytes sectors per track (= 32 sectors of 512 bytes)
        .DB 5   ;BSH - block shift factor
        .DB 31  ;BLM - block mask
        .DB 1   ;EXM - Extent mask
        .DW 2043 ; (2047-4) DSM - Storage size (blocks - 1)
        .DW 511 ;DRM - Number of directory entries - 1
        .DB 240 ;AL0 - 1 bit set per directory block
        .DB 0   ;AL1 -            "
        .DW 0   ;CKS - DIR check vector size (DRM+1)/4 (0=fixed disk)
        .DW 1   ;OFF - Reserved tracks

dpb
        .DW 128 ;SPT - 128 bytes sectors per track (= 32 sectors of 512 bytes)
        .DB 5   ;BSH - block shift factor
        .DB 31  ;BLM - block mask
        .DB 1   ;EXM - Extent mask
        .DW 2047 ;DSM - Storage size (blocks - 1)
        .DW 511 ;DRM - Number of directory entries - 1
        .DB 240 ;AL0 - 1 bit set per directory block
        .DB 0   ;AL1 -            "
        .DW 0   ;CKS - DIR check vector size (DRM+1)/4 (0=fixed disk)
        .DW 0   ;OFF - Reserved tracks


; =========================================================================== ;
; BOOT                                                                        ;
; =========================================================================== ;
; The BOOT entry point gets control from the cold start loader and is         ;
; responsible for basic system initialization, including sending a sign-on    ;
; message, which can be omitted in the first version.                         ;
; If the IOBYTE function is implemented, it must be set at this point.        ;
; The various system parameters that are set by the WBOOT entry point must be ;
; initialized, and control is transferred to the CCP at 3400 + b for further  ;
; processing. Note that register C must be set to zero to select drive A.     ;
; =========================================================================== ;
BOOT_
    xra     a               ;xor     a
    sta     IOBYT           ;ld      (IOBYT), a      ; Clear IOBYTE
    sta     CDISK           ;ld      (CDISK), a      ; Select Disk 0 & User 0
    lxi     h, BiosMsg      ;ld      hl,BiosMsg      ; Print a message
    call    puts
    ;
    ; Set up the execution of AUTOEXEC.SUB if required
    ;
                            ;ld      a, SYSFLAG_OPC  ; A = SYSFLAG opcode
                            ;out     (STO_OPCD), a   ; Write the opcode
    in      SYSFLAGS        ;in      a, (EXC_RD_OPCD); Check if AUTOEXEC execution is requierd
    ani     $01             ;and     $01             ; Isolate AUTOEXEC flag
    jz      GOCPM           ;jr      z, GOCPM        ; Jump if flag = 0 (nothing to set up)
    lxi     b, CCPAuto      ;ld      bc, CCPAuto     ; Flag = 1, BC = address of AUTOEXEC command string
    lxi     h, CCPFIRS      ;ld      hl, CCPFIRS     ; HL = address of the first char of CCP input string
bufCopy     ; Copy the AUTOEXEC command string into the CCP input buffer
    ldax    b               ;ld      a, (bc)         ; A = current command string char
    cpi     eos             ;cp      eos             ; End of string reached?
    jz      bufCopyEnd      ;jr      z, bufCopyEnd   ; Yes, jump
    mov     m, a            ;ld      (hl), a         ; No, load it the CCP input buffer
    inx     b               ;inc     bc              ; Increment command string character pointer
    inx     h               ;inc     hl              ; Increment CCP input buffer character pointer
    jmp     bufCopy         ;jr      bufCopy         ; Copy  next character
bufCopyEnd  ; Calculate command string lenght and store it to CCP input buffer 
            ; lenght variable
    lxi     b, CCPFIRS      ;ld      bc, CCPFIRS     ; BC = address of the first char of CCP input string
                            ;xor     a               ; Carry = 0
                            ;sbc     hl, bc          ; L = command string lenght (H = 0 always)
                            ;ld      a, l            ; A = command string lenght
    mov     a, l
    sub     c               ; A = L - C = command string lenght (always < 255)
    sta     CCPLEN          ;ld      (CCPLEN), a     ; Store it into CCP buffer lenght variable
    jmp     GOCPM           ;jr      GOCPM

; =========================================================================== ;
; WBOOT                                                                       ;
; =========================================================================== ;
; The WBOOT entry point gets control when a warm start occurs.                ;
; A warm start is performed whenever a user program branches to location      ;
; 0000H, or when the CPU is reset from the front panel. The CP/M system must  ;
; be loaded from the first two tracks of drive A up to, but not including,    ;
; the BIOS, or CBIOS, if the user has completed the patch. System parameters  ;
; must be initialized as follows:                                             ;
;                                                                             ;
; location 0,1,2                                                              ;
;     Set to JMP WBOOT for warm starts (000H: JMP 4A03H + b)                  ;
;                                                                             ;
; location 3                                                                  ;
;     Set initial value of IOBYTE, if implemented in the CBIOS                ;
;                                                                             ;
; location 4                                                                  ;
;     High nibble = current user number, low nibble = current drive           ;
;                                                                             ;
; location 5,6,7                                                              ;
;     Set to JMP BDOS, which is the primary entry point to CP/M for transient ;
;     programs. (0005H: JMP 3C06H + b)                                        ;
;                                                                             ;
; Refer to Section 6.9 for complete details of page zero use. Upon completion ;
; of the initialization, the WBOOT program must branch to the CCP at 3400H+b  ;
; to restart the system.                                                      ;
; Upon entry to the CCP, register C is set to the drive to select after system;
; initialization. The WBOOT routine should read location 4 in memory, verify  ;
; that is a legal drive, and pass it to the CCP in register C.                ;
; =========================================================================== ;

WBOOT_
    ;
    ; Load CCP+BDOS from the system area of disk 0 (track 0).
    ; CCP+BDOS are stored from host sector 0 to 10. 
    ; Total 11 host sectors (each 512 bytes large) to load
    ;
    lxi     sp, $80         ;ld      sp, $80         ; Use space below buffer for stack 
    lxi     h, WbootMSG     ;ld      hl, WbootMSG    ; Print a message
    call    puts
    ; Select host disk 0 and reset CONST mode
    mvi     a, SELDISK_OPC  ;ld      a, SELDISK_OPC  ; Select SELDISK opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    xra     a               ;xor     a               ; disk = 0
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a; Select it
    ; Select host track 0
    mvi     a, SELTRCK_OPC  ;ld      a, SELTRCK_OPC  ; Select SELTRACK opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    xra     a               ;xor     a               ; track = 0
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a; LSB track = 0
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a; MSB track = 0
    ; Setup load 
    mvi     d, 0            ;ld      d, 0            ; D = first host sector to load
    mvi     e, 11           ;ld      e, 11           ; E = host sectors to load (= 11 * 512 bytes)
    lxi     h, CCP          ;ld      hl, CCP         ; HL = CCP starting address (destination)
WBTLOOP
    ; Select current host sector
    mvi     a, SELSECT_OPC  ;ld      a, SELSECT_OPC  ; Select SELSECT opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    mov     a, d            ;ld      a, d            ; A = current host sector number
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a; Select it
    ; Read current host sector (512 bytes) into (HL)
                            ;ld      c, EXC_RD_OPCD  ; Set the EXECUTE READ OPCODE port into C
    mvi     a, RDSECT_OPC   ;ld      a, RDSECT_OPC   ; Select READSECT opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    mvi     b, 0            ;ld      b, 0            ; Byte counter = 256
                            ;inir                    ; Read 256 byte to hstbuf
inir_1
    in      EXC_RD_OPCD
    mov     m, a
    inx     h
    dcr     b
    jnz     inir_1
                            ;inir                    ; Read 256 byte to hstbuf
inir_2
    in      EXC_RD_OPCD
    mov     m, a
    inx     h
    dcr     b
    jnz     inir_2
    ; Check read errors
    mvi     a, ERRDSK_OPC   ;ld      a, ERRDSK_OPC  ; Select ERRDISK opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    in      EXC_RD_OPCD     ;in      a, (EXC_RD_OPCD); A = read error code
    ora     a               ;or      a               ; Set flags
    jnz     FATALERR        ;jr      nz, FATALERR    ; Ia A > 0 jump to error print
    ; All done?
    inr     d               ;inc     d               ; Increment for next sector
    dcr     e               ;dec     e               ; Decrement counter E
    jnz     WBTLOOP         ;jr      nz, WBTLOOP     ; Jump if E > 0
    jmp     WBTEND          ;jr      WBTEND          ; All done (E = 0), jump to the CCP

FATALERR
    lxi     h, FatalMsg     ;ld      hl, FatalMsg    ; Print an error message
    call    puts
    hlt                     ;halt

WBTEND
    lxi     h, CRLFLF       ;ld      hl, CRLFLF      ; Print a CR with two LF
    call    puts

GOCPM
    ; Init code for blocking and deblocking algorithm
    ; (see CP/M 2.2 Alteration Guide p.34 and APPENDIX G)
    xra     a               ;xor a                   ; 0 to accumulator
    sta     hstact          ;ld  (hstact),a          ; Host buffer inactive
    sta     unacnt          ;ld  (unacnt),a          ; Clear unalloc count
    ; CP/M init
    lxi     h, CPMMsg       ;ld      hl, CPMMsg      ; Print a message
    call    puts
    mvi     a, $C3          ;ld      a, $C3          ; C3 IS A JMP INSTRUCTION
    sta     $0000           ;ld      ($0000), a      ; FOR JMP TO WBOOT
    lxi     h, WBOOTE       ;ld      hl,WBOOTE       ; WBOOT ENTRY POINT
    shld    $0001           ;ld      ($0001), hl     ; SET ADDRESS FIELD FOR JMP AT 0
    sta     $0005           ;ld      ($0005), a      ; FOR JMP TO BDOS
    lxi     h, BDOS         ;ld      hl, BDOS        ; BDOS ENTRY POINT
    shld    $0006           ;ld      ($0006), hl     ; ADDRESS FIELD OF JUMP AT 5 TO BDOS
    lxi     b, $0080        ;ld      bc, $0080       ; DEFAULT DMA ADDRESS IS 80H
    call    SETDMA_
    ;ei                     ; DO NOT ENABLE THE INTERRUPT SYSTEM
    ; Check if current disk is valid
    lda     CDISK           ;ld      a, (CDISK)      ; GET CURRENT USER/DISK NUMBER (UUUUDDDD)
    ani     $0f             ;and     $0f             ; Isolate the disk number.
    cpi     NDISKS          ;cp      NDISKS          ; Drive number ok?
    jc      WBTDSKOK        ;jr      c, WBTDSKOK     ; Yes, jump (Carry set if A < NDISKS)
    lda     CDISK           ;ld      a, (CDISK)      ; No, set disk 0 (previous user)
    ani     $f0             ;and     $f0
    sta     CDISK           ;ld      (CDISK), a      ; Save User/Disk    
WBTDSKOK    
    lda     CDISK           ;ld      a, (CDISK)
    mov     c, a            ;ld      c, a            ; C = current User/Disk for CCP jump (UUUUDDDD)
    jmp     CCP             ;jp      CCP             ; GO TO CCP FOR FURTHER PROCESSING

; =========================================================================== ;
; CONST                                                                       ;
; =========================================================================== ;
; You should sample the status of the currently assigned console device and   ;
; return 0FFH in register A if a character is ready to read and 00H in        ;
; register A if no console characters are ready.                              ;
; =========================================================================== ;
CONST_
    in      SYSFLAGS        ; Read SYSFLAGS
    ani     $04             ; Isolate D2 (Serial RX char available flag)
    rz                      ; Return with A = 0 if no char is present
    mvi     a, $ff          ; Otherwise A = $ff
    ret

; =========================================================================== ;
; CONIN                                                                       ;
; =========================================================================== ;
; The next console character is read into register A, and the parity bit is   ;
; set, high-order bit, to zero. If no console character is ready, wait until  ;
; a character is typed before returning.                                      ;
; =========================================================================== ;
CONIN_
    in      SYSFLAGS        ; Read SYSFLAGS
    ani     $04             ; Isolate D2 (Serial RX char available flag)
    jz      CONIN_          ; Wait until a char is present
    in      SERIAL_RX       ; Get the char
    ret

; =========================================================================== ;
; CONOUT                                                                      ;
; =========================================================================== ;
; The character is sent from register C to the console output device.         ;
; The character is in ASCII, with high-order parity bit set to zero. You      ;
; might want to include a time-out on a line-feed or carriage return, if the  ;
; console device requires some time interval at the end of the line (such as  ;
; a TI Silent 700 terminal). You can filter out control characters that cause ;
; the console device to react in a strange way (CTRL-Z causes the Lear-       ;
; Siegler terminal to clear the screen, for example).                         ;
; =========================================================================== ;
CONOUT_
    mvi     a, SERTX_OPC    ;ld      a, SERTX_OPC    ; A = SERIAL TX opcode
    out     STO_OPCD        ;out     (STO_OPCD), a   ; Write the opcode
    mov     a, c            ;ld      a, c
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a; Send A to serial Tx
    ret

; =========================================================================== ;
; LIST                                                                        ;
; =========================================================================== ;
; The character is sent from register C to the currently assigned listing     ;
; device. The character is in ASCII with zero parity bit.                     ;
; =========================================================================== ;
LIST_
    ret                     ; Not implemented

; =========================================================================== ;
; PUNCH                                                                       ;
; =========================================================================== ;
; The character is sent from register C to the currently assigned punch       ;
; device. The character is in ASCII with zero parity.                         ;
; =========================================================================== ;
PUNCH_
    ret                     ; Not implemented

; =========================================================================== ;
; READER                                                                      ;
; =========================================================================== ;
; The next character is read from the currently assigned reader device into   ;
; register A with zero parity (high-order bit must be zero); an end-of-file   ;
; condition is reported by returning an ASCII CTRL-Z(1AH).                    ;
; =========================================================================== ;
READER_
    mvi     a, $1a          ;ld      a, $1a          ; Enter an EOF for now (READER not implemented)
    ret

; =========================================================================== ;
; HOME                                                                        ;
; =========================================================================== ;
; The disk head of the currently selected disk (initially disk A) is moved to ;
; the track 00 position. If the controller allows access to the track 0 flag  ;
; from the drive, the head is stepped until the track 0 flag is detected. If  ;
; the controller does not support this feature, the HOME call is translated   ;
; into a call to SETTRK with a parameter of 0.                                ;
; =========================================================================== ;
HOME_
    lxi     b, 0            ;ld      bc, 0
    jmp     SETTRK_         ;jp      SETTRK_

; =========================================================================== ;
; SELDSK                                                                      ;
; =========================================================================== ;
; The disk drive given by register C is selected for further operations,      ;
; where register C contains 0 for drive A, 1 for drive B, and so on up to 15  ;
; for drive P (the standard CP/M distribution version supports four drives).  ;
; On each disk select, SELDSK must return in HL the base address of a 16-byte ;
; area, called the Disk Parameter Header, described in Section 6.10.          ;
; For standard floppy disk drives, the contents of the header and associated  ;
; tables do not change; thus, the program segment included in the sample      ;
; CBIOS performs this operation automatically.                                ;
;                                                                             ;
; If there is an attempt to select a nonexistent drive, SELDSK returns        ;
; HL = 0000H as an error indicator. Although SELDSK must return the header    ;
; address on each call, it is advisable to postpone the physical disk select  ;
; operation until an I/O function (seek, read, or write) is actually          ;
; performed, because disk selects often occur without ultimately performing   ;
; any disk I/O, and many controllers unload the head of the current disk      ;
; before selecting the new drive. This causes an excessive amount of noise    ;
; and disk wear. The least significant bit of register E is zero if this is   ;
; the first occurrence of the drive select since the last cold or warm start. ;
; =========================================================================== ;
SELDSK_
    lxi     h, $0000        ;ld      hl, $0000       ; HL = error code
    mov     a, c            ;ld      a, c            ; A = disk number
    cpi     NDISKS          ;cp      NDISKS          ; Drive number ok?
    jc      calcHL          ;jr      c, calcHL       ; Yes, jump
    xra     a               ;xor a                   ; No, set disk 0 as current disk
    sta     CDISK           ;ld      (CDISK), a      ; Save disk 0, user 0
    ret
    
calcHL
    ; Code for blocking and deblocking algorithm
    ; (see CP/M 2.2 Alteration Guide p.34 and APPENDIX G)
    sta     sekdsk          ;LD      (sekdsk),A
    mov     l, a            ; l = disk number 0,1,2,3,...
                            ;RLC     a               ; *2
                            ;RLC     a               ; *4
                            ;RLC     a               ; *8
                            ;RLC     a               ; *16
                            ;LD      HL,dpbase
                            ;LD      b,0
                            ;LD      c,A 
                            ;ADD     HL,BC
    dad     h               ; *2
    dad     h               ; *4
    dad     h               ; *8
    dad     h               ; *16
    lxi     d, dpbase
    dad     d               ; hl = dpbase+(diskno*16)
    ret

; =========================================================================== ;
; SETTRK                                                                      ;
; =========================================================================== ;
; Register BC contains the track number for subsequent disk accesses on the   ;
; currently selected drive. The sector number in BC is the same as the number ;
; returned from the SECTRN entry point. You can choose to seek the selected   ;
; track at this time or delay the seek until the next read or write actually  ;
; occurs. Register BC can take on values in the range 0-76 corresponding to   ;
; valid track numbers for standard floppy disk drives and 0-65535 for         ;
; nonstandard disk subsystems.                                                ;
; =========================================================================== ;
SETTRK_  
                            ;ld      (sektrk), bc    ; Set track passed from BDOS in register BC.
    mov     h, b
    mov     l, c
    shld    sektrk
    ret

; =========================================================================== ;
; SETSEC                                                                      ;
; =========================================================================== ;
; Register BC contains the sector number, 1 through 26, for subsequent disk   ;
; accesses on the currently selected drive. The sector number in BC is the    ;
; same as the number returned from the SECTRAN entry point. You can choose to ;
; send this information to the controller at this point or delay sector       ;
; selection until a read or write operation occurs.                           ;
; =========================================================================== ;
SETSEC_
                                ;ld      (seksec), bc    ; Set sector passed from BDOS in register BC.
    mov     h, b
    mov     l, c
    shld    seksec
    ret

; =========================================================================== ;
; SETDMA                                                                      ;
; =========================================================================== ;
; Register BC contains the DMA (Disk Memory Access) address for subsequent    ;
; read or write operations. For example, if B = 00H and C = 80H when SETDMA   ;
; is called, all subsequent read operations read their data into 80H through  ;
; 0FFH and all subsequent write operations get their data from 80H through    ;
; 0FFH, until the next call to SETDMA occurs. The initial DMA address is      ;
; assumed to be 80H. The controller need not actually support Direct Memory   ;
; Access. If, for example, all data transfers are through I/O ports, the      ;
; CBIOS that is constructed uses the 128 byte area starting at the selected   ;
; DMA address for the memory buffer during the subsequent read or write       ;
; operations.                                                                 ;
; =========================================================================== ;
SETDMA_
                                ;ld      (dmaAddr), bc   ; Save the DMA (Disk Memory Access) address
    mov     h, b
    mov     l, c
    shld    dmaAddr
    ret

; =========================================================================== ;
; READ                                                                        ;
; =========================================================================== ;
; Assuming the drive has been selected, the track has been set, and the DMA   ;
; address has been specified, the READ subroutine attempts to read one sector ;
; based upon these parameters and returns the following error codes in        ;
; register A:                                                                 ;
;                                                                             ;
;     0 - no errors occurred                                                  ;
;     1 - non recoverable error condition occurred                            ;
;                                                                             ;
; Currently, CP/M responds only to a zero or nonzero value as the return      ;
; code. That is, if the value in register A is 0, CP/M assumes that the disk  ;
; operation was completed properly. If an error occurs the CBIOS should       ;
; attempt at least 10 retries to see if the error is recoverable. When an     ;
; error is reported the BDOS prints the message BDOS ERR ON x: BAD SECTOR.    ;
; The operator then has the option of pressing a carriage return to ignore    ;
; the error, or CTRL-C to abort.                                              ;
; =========================================================================== ;
READ_
    ;
    ; Code for blocking and deblocking algorithm
    ; (see CP/M 2.2 Alteration Guide p.34 and APPENDIX G)
    ;
    xra     a               ;xor a
    sta     unacnt          ;ld      (unacnt), a
    mvi     a, 1            ;ld      a, 1
    sta     readop          ;ld      (readop), a     ; read operation
    sta     rsflag          ;ld      (rsflag), a     ; must read data
    mvi     a, wrual        ;ld      a, wrual
    sta     wrtype          ;ld      (wrtype), a     ; treat as unalloc
    jmp     rwoper          ;jp      rwoper          ; to perform the read

; =========================================================================== ;
; WRITE                                                                       ;
; =========================================================================== ;
; Data is written from the currently selected DMA address to the currently    ;
; selected drive, track, and sector. For floppy disks, the data should be     ;
; marked as nondeleted data to maintain compatibility with other CP/M         ;
; systems. The error codes given in the READ command are returned in register ;
; A, with error recovery attempts as described above.                         ;
; =========================================================================== ;
WRITE_
    ;
    ; Code for blocking and deblocking algorithm
    ; (see CP/M 2.2 Alteration Guide p.34 and APPENDIX G)
    ;
    xra     a               ;xor     a               ; 0 to accumulator
    sta     readop          ;ld      (readop), a     ; not a read operation
    mov     a, c            ;ld      a, c            ; write type in c
    sta     wrtype          ;ld      (wrtype), a
    cpi     wrual           ;cp      wrual           ; write unallocated?
    jnz     chkuna          ;jr      nz, chkuna      ; check for unalloc
    ; write to unallocated, set parameters
    mvi     a, blksiz/128   ;ld      a, blksiz/128   ; next unalloc recs
    sta     unacnt          ;ld      (unacnt), a
    lda     sekdsk          ;ld      a, (sekdsk)     ; disk to seek
    sta     unadsk          ;ld      (unadsk), a     ; unadsk = sekdsk
    lhld    sektrk          ;ld      hl, (sektrk)
    shld    unatrk          ;ld      (unatrk), hl    ; unatrk = sectrk
    lda     seksec          ;ld      a, (seksec)
    sta     unasec          ;ld      (unasec), a     ; unasec = seksec
chkuna:
    ; check for write to unallocated sector
    lda     unacnt          ;ld      a,(unacnt)      ; any unalloc remain?
    ora     a               ;or      a   
    jz      alloc           ;jr     z, alloc        ; skip if not
    ; more unallocated records remain
    dcr     a               ;dec     a               ; unacnt = unacnt-1
    sta     unacnt          ;ld      (unacnt), a
    lda     sekdsk          ;ld      a, (sekdsk)     ; same disk?
    lxi     h, unadsk       ;ld      hl, unadsk
    cmp     m               ;cp      (hl)            ; sekdsk = unadsk?
    jnz     alloc           ;jp      nz, alloc       ; skip if not
    ; disks are the same
    lxi     h, unatrk       ;ld      hl, unatrk
    call    sektrkcmp       ; sektrk = unatrk?
    jnz     alloc           ;jp      nz, alloc       ; skip if not
    ;   tracks are the same
    lda     seksec          ;ld      a, (seksec)     ; same sector?
    lxi     h, unasec       ;ld      hl, unasec
    cmp     m               ;cp      (hl)            ; seksec = unasec?
    jnz     alloc           ;jp      nz, alloc       ; skip if not
    ; match, move to next sector for future ref
    inr     m               ;inc     (hl)            ; unasec = unasec+1
    mov     a, m            ;ld      a, (hl)         ; end of track?
    cpi     cpmspt          ;cp      cpmspt          ; count CP/M sectors
    jc      noovf           ;jr      c, noovf        ; skip if no overflow
    ; overflow to next track
    mvi     m, 0            ;ld      (hl), 0         ; unasec = 0
    lhld    unatrk          ;ld      hl, (unatrk)
    inx     h               ;inc     hl
    shld    unatrk          ;ld      (unatrk), hl    ; unatrk = unatrk+1
noovf:
    ; match found, mark as unnecessary read
    xra     a               ;xor     a               ; 0 to accumulator
    sta     rsflag          ;ld      (rsflag), a     ; rsflag = 0
    jmp     rwoper          ;jr      rwoper          ; to perform the write
alloc:
    ; not an unallocated record, requires pre-read
    xra     a               ;xor     a               ; 0 to accum
    sta     unacnt          ;ld      (unacnt), a     ; unacnt = 0
    inr     a               ;inc     a               ; 1 to accum
    sta     rsflag          ;ld      (rsflag), a     ; rsflag = 1
rwoper:
    ; enter here to perform the read/write
    xra     a               ;xor     a               ; zero to accum
    sta     erflag          ;ld      (erflag), a     ; no errors (yet)
    lda     seksec          ;ld      a, (seksec)     ; compute host sector
    ora     a               ;or      a               ; carry = 0
    rar                     ;rra                     ; shift right
    ora     a               ;or      a               ; carry = 0
    rar                     ;rra                     ; shift right
    sta     sekhst          ;ld      (sekhst), a     ; host sector to seek
    ; active host sector?
    lxi     h, hstact       ;ld      hl, hstact      ; host active flag
    mov     a, m            ;ld      a, (hl)
    mvi     m, 1            ;ld      (hl), 1         ; always becomes 1
    ora     a               ;or      a               ; was it already?
    jz      filhst          ;jr      z, filhst       ; fill host if not
    ; host buffer active, same as seek buffer?
    lda     sekdsk          ;ld      a, (sekdsk)
    lxi     h, hstdsk       ;ld      hl, hstdsk      ; same disk?
    cmp     m               ;cp      (hl)            ; sekdsk = hstdsk?
    jnz     nomatch         ;jr      nz, nomatch
    ; same disk, same track?
    lxi     h, hsttrk       ;ld      hl, hsttrk
    call    sektrkcmp       ; sektrk = hsttrk?
    jnz     nomatch         ;jr      nz, nomatch
    ; same disk, same track, same buffer?
    lda     sekhst          ;ld      a, (sekhst)
    lxi     h, hstsec       ;ld      hl, hstsec      ; sekhst = hstsec?
    cmp     m               ;cp      (hl)
    jz      match           ;jr      z, match        ; skip if match
nomatch:
    ; proper disk, but not correct sector
    lda     hstwrt          ;ld      a, (hstwrt)     ; host written?
    ora     a               ;or      a
    cnz     writehst        ;call    nz, writehst    ; clear host buff
filhst:
    ; may have to fill the host buffer
    lda     sekdsk          ;ld      a, (sekdsk)
    sta     hstdsk          ;ld      (hstdsk), a
    lhld    sektrk          ;ld      hl, (sektrk)
    shld    hsttrk          ;ld      (hsttrk), hl
    lda     sekhst          ;ld      a, (sekhst)
    sta     hstsec          ;ld      (hstsec), a
    lda     rsflag          ;ld      a, (rsflag)     ; need to read?
    ora     a               ;or      a
    cnz     readhst         ;call    nz, readhst     ; yes, if 1
    xra     a               ;xor     a               ; 0 to accum
    sta     hstwrt          ;ld      (hstwrt), a     ; no pending write
match:
    ; copy data to or from buffer
    lda     seksec          ;ld      a,  (seksec)    ; mask buffer number
    ani     secmsk          ;and     secmsk          ; least signif bits
    mov     l, a            ;ld      l,  a           ; ready to shift
    mvi     h, 0            ;ld      h,  0           ; double count
    dad     h               ;add     hl, hl
    dad     h               ;add     hl, hl
    dad     h               ;add     hl, hl
    dad     h               ;add     hl, hl
    dad     h               ;add     hl, hl
    dad     h               ;add     hl, hl
    dad     h               ;add     hl, hl
    ; hl has relative host buffer address
    lxi     d, hstbuf       ;ld      de, hstbuf
    dad     d               ;add     hl, de          ; hl = host address
    xchg                    ;ex      de, hl          ; now in DE
    lhld    dmaAddr         ;ld      hl, (dmaAddr)   ; get/put CP/M data
    mvi     c, 128          ;ld      c, 128          ; length of move
    lda     readop          ;ld      a, (readop)     ; which way?
    ora     a               ;or      a
    jnz     rwmove          ;jr      nz, rwmove      ; skip if read
    ; write operation, mark and switch direction
    mvi     a, 1            ;ld      a, 1
    sta     hstwrt          ;ld      (hstwrt), a     ; hstwrt = 1
    xchg                    ;ex      de, hl          ; source/dest swap
rwmove:
    ; C initially 128, DE is source, HL is dest
    ldax    d               ;ld      a, (de)         ; source character
    inx     d               ;inc     de
    mov     m, a            ;ld      (hl), a         ; to dest
    inx     h               ;inc     hl
    dcr     c               ;dec     c               ; loop 128 times
    jnz     rwmove          ;jr      nz, rwmove
    ; data has been moved to/from host buffer
    lda     wrtype          ;ld      a, (wrtype)     ; write type
    cpi     wrdir           ;cp      wrdir           ; to directory?
    lda     erflag          ;ld      a, (erflag)     ; in case of errors
    rnz                     ;ret     nz              ; no further processing
    ; clear host buffer for directory write
    ora     a               ;or      a               ; errors?
    rnz                     ;ret     nz              ; skip if so
    xra     a               ;xor     a               ; 0 to accum
    sta     hstwrt          ;ld      (hstwrt), a     ; buffer written
    call    writehst
    lda     erflag          ;ld      a, (erflag)
    ret

; Utility subroutine for 16-bit compare
sektrkcmp:
    ; HL = .unatrk or .hsttrk, compare with sektrk
    xchg                    ;ex      de, hl
    lxi     h, sektrk       ;ld      hl, sektrk
    ldax    d               ;ld      a, (de)         ; low byte compare
    cmp     m               ;cp      (HL)            ; same?
    rnz                     ;ret     nz              ; return if not
    ; low bytes equal, test high 1s
    inx     d               ;inc     de
    inx     h               ;inc     hl
    ldax    d               ;ld      a, (de)
    cmp     m               ;cp      (hl)            ; sets flags
    ret

;------------------------------------------------------------------------------------------------
; Read physical sector from host
;------------------------------------------------------------------------------------------------

readhst:
    push    b               ;push    bc
    push    h               ;push    hl
    call    setDTS          ; Select disk, track, sector
    ; Read current host sector (512 byte) to hstbuf
                            ;ld      c, EXC_RD_OPCD  ; Set the EXECUTE READ OPCODE port into C
    lxi     h, hstbuf       ;ld      hl, hstbuf      ; HL points to buffer hstbuf
    mvi     a, RDSECT_OPC   ;ld      a, RDSECT_OPC   ; Select READSECT opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    mvi     b, 0            ;ld      b, 0            ; Byte counter = 256
                            ;inir                    ; Read 256 byte to hstbuf
inir_3
    in      EXC_RD_OPCD
    mov     m, a
    inx     h
    dcr     b
    jnz     inir_3
                            ;inir                    ; Read 256 byte to hstbuf
inir_4
    in      EXC_RD_OPCD
    mov     m, a
    inx     h
    dcr     b
    jnz     inir_4
    ; Check for errors
    mvi     a, ERRDSK_OPC   ;ld      a, ERRDSK_OPC      ; Select ERRDISK opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    in      EXC_RD_OPCD     ;in      a, (EXC_RD_OPCD); Read error code into A...
    sta     erflag          ;ld     (erflag), a      ; ... and store it into erflag
    pop     h               ;pop     hl
    pop     b               ;pop     bc
    ret

; Set disk, track and sector routine for a read or write operation
setDTS
    ; Select hstdsk host disk
    lda     lastDsk         ;ld      a, (lastDsk)
    mov     b, a            ;ld      b, a            ; B = last disk number
    lda     hstdsk          ;ld      a, (hstdsk)     ; A = new disk number
    cmp     b               ;cp      b               ; previous disk = new disk?
    jz      setTrack        ;jr      z, setTrack     ; Yes, jump ahead
    mvi     a, SELDISK_OPC  ;ld      a, SELDISK_OPC  ; No, Select SELDISK opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    lda     hstdsk          ;ld      a, (hstdsk)     ; Select the disk number (hstdsk)
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a
    sta     lastDsk         ;ld      (lastDsk), a    ; Update last disk number
    ; Select hsttrk host track
setTrack
    mvi     a, SELTRCK_OPC  ;ld      a, SELTRCK_OPC  ; Select SELTRACK opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    lda     hsttrk          ;ld      a, (hsttrk)     ; Select the track number (hsttrk) LSB
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a
    lda     hsttrk+1        ;ld      a, (hsttrk+1)   ; Select the track number (hsttrk) MSB
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a
    ; Select hstsec host sector
    mvi     a, SELSECT_OPC  ;ld      a, SELSECT_OPC  ; Select SELSECT opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    lda     hstsec          ;ld      a, (hstsec)     ; Select the sector number (hstsec)
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a
    ret
    
lastDsk     .fill   1       ; Last disk number (= $ff after cold boot)

;------------------------------------------------------------------------------------------------
; Write physical sector to host
;------------------------------------------------------------------------------------------------

writehst
    push    b               ;push    bc
    push    h               ;push    hl
    call    setDTS          ; Select disk, track, sector
    ; Write current host sector (512 byte) to hstbuf
                            ;ld      c, EXC_WR_OPCD  ; Set the EXECUTE WRITE OPCODE port into C
    lxi     h, hstbuf       ;ld      hl, hstbuf      ; HL points to buffer hstbuf
    mvi     a, WRTSECT_OPC  ;ld      a, WRTSECT_OPC  ; Select WRITESECT opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    mvi     b, 0            ;ld      b, 0            ; Byte counter = 256
; ************************************************************************************* CONTINUARE DA QUI IN GIU'
    ;                       otir                    ; Write 256 byte to hstbuf
otir_1
    mov     a, m
    out     EXC_WR_OPCD
    inx     h
    dcr     b
    jnz     otir_1
                            ;otir                    ; Write 256 byte to hstbuf
otir_2
    mov     a, m
    out     EXC_WR_OPCD
    inx     h
    dcr     b
    jnz     otir_2
    ; Check for errors
    mvi     a, ERRDSK_OPC   ;ld      a, ERRDSK_OPC      ; Select ERRDISK opcode (IOS)
    out     STO_OPCD        ;out     (STO_OPCD), a
    in      EXC_RD_OPCD     ;in      a, (EXC_RD_OPCD); Read error code into A...
    sta     erflag          ;ld  (erflag),a          ; ... and store it into erflag
    pop     h               ;pop     hl
    pop     b               ;pop     bc
    ret

; =========================================================================== ;
; LISTST                                                                      ;
; =========================================================================== ;
; You return the ready status of the list device used by the DESPOOL program  ;
; to improve console response during its operation. The value 00 is returned  ;
; in A if the list device is not ready to accept a character and 0FFH if a    ;
; character can be sent to the printer. A 00 value should be returned if LIST ;
; status is not implemented.                                                  ;
; =========================================================================== ;
LISTST_
    xra     a               ;xor     a               ; A = 0 (not implemented)
    ret

; =========================================================================== ;
; SECTRAN                                                                     ;
; =========================================================================== ;
; Logical-to-physical sector translation is performed to improve the overall  ;
; response of CP/M. Standard CP/M systems are shipped with a skew factor of   ;
; 6, where six physical sectors are skipped between each logical read         ;
; operation. This skew factor allows enough time between sectors for most     ;
; programs to load their buffers without missing the next sector. In          ;
; particular computer systems that use fast processors, memory, and disk      ;
; subsystems, the skew factor might be changed to improve overall response.   ;
; However, the user should maintain a single-density IBM-compatible version   ;
; of CP/M for information transfer into and out of the computer system, using ;
; a skew factor of 6.                                                         ;
;                                                                             ;
; In general, SECTRAN receives a logical sector number relative to zero in BC ;
; and a translate table address in DE. The sector number is used as an index  ;
; into the translate table, with the resulting physical sector number in HL.  ;
; For standard systems, the table and indexing code is provided in the CBIOS  ;
; and need not be changed.                                                    ;
; =========================================================================== ;
SECTRN_
    mov     h, b            ;ld      h, b            ; HL = BC
    mov     l, c            ;ld      l, c
    ret

; =========================================================================== ;
;
; Send a string to the serial line, HL contains the pointer to the string.
; NOTE: Only A and HL are used
;
; =========================================================================== ;
puts
    mov     a, m            ;ld      a, (hl)         ; A = current char to print
    cpi     eos             ;cp      eos             ; End of string reached?
    jz      puts_end        ;jr      z, puts_end     ; Yes, jump
    mvi     a, SERTX_OPC    ;ld      a, SERTX_OPC    ; A = SERIAL TX opcode
    out     STO_OPCD        ;out     (STO_OPCD), a   ; Write the opcode
    mov     a, m            ;ld      a, (hl)         ; A = current char to print
    out     EXC_WR_OPCD     ;out     (EXC_WR_OPCD), a; Print A
    inx     h               ;inc     hl              ; Increment character pointer
    jmp     puts            ;jr      puts            ; Transmit next character

puts_end
    ret
    
; =========================================================================== ;
;
; MESSAGES
;
; =========================================================================== ;

BiosMsg     .db     cr, lf, lf, "V20-MBC CP/M 2.2 (8080) BIOS - S140520", cr, lf, eos ;
CRLFLF      .db     cr, lf, lf, eos
CPMMsg      .db     "CP/M 2.2 Copyright 1979 (c) by Digital Research", cr, lf, eos
FatalMsg    .db     cr, lf, "FATAL DISK READ ERROR - SYSTEM HALTED", eos
WbootMSG    .db     cr, lf, "CP/M WARM BOOT...", cr, lf, eos
CCPAuto     .db     "SUBMIT AUTOEXEC", eos

; =========================================================================== ;
; THE REMAINDER OF THE CBIOS IS RESERVED UNINITIALIZED DATA AREA, AND DOES    ;
; NOT NEED TO BE A PART OF THE SYSTEM MEMORY IMAGE (THE SPACE MUST BE         ;
; AVAILABLE, HOWEVER).                                                        ;
; =========================================================================== ;

dirbuf:     .ds 128         ; scratch directory area
alv00:      .ds 257         ; allocation vector 0
alv01:      .ds 257         ; allocation vector 1
alv02:      .ds 257         ; allocation vector 2
alv03:      .ds 257         ; allocation vector 3
alv04:      .ds 257         ; allocation vector 4
alv05:      .ds 257         ; allocation vector 5
alv06:      .ds 257         ; allocation vector 6
alv07:      .ds 257         ; allocation vector 7
alv08:      .ds 257         ; allocation vector 8
alv09:      .ds 257         ; allocation vector 9
alv10:      .ds 257         ; allocation vector 10
alv11:      .ds 257         ; allocation vector 11
alv12:      .ds 257         ; allocation vector 12
alv13:      .ds 257         ; allocation vector 13
alv14:      .ds 257         ; allocation vector 14
alv15:      .ds 257         ; allocation vector 15

;
; Blocking and deblocking algorithm variables
; (see CP/M 2.2 Alteration Guide p.34 and APPENDIX G)
;

sekdsk:     .ds 1           ; seek disk number
sektrk:     .ds 2           ; seek track number
seksec:     .ds 2           ; seek sector number
;
hstdsk:     .ds 1           ; host disk number
hsttrk:     .ds 2           ; host track number
hstsec:     .ds 1           ; host sector number
;
sekhst:     .ds 1           ; seek shr secshf
hstact:     .ds 1           ; host active flag
hstwrt:     .ds 1           ; host written flag
;
unacnt:     .ds 1           ; unalloc rec cnt
unadsk:     .ds 1           ; last unalloc disk
unatrk:     .ds 2           ; last unalloc track
unasec:     .ds 1           ; last unalloc sector
;
erflag:     .ds 1           ; error reporting
rsflag:     .ds 1           ; read sector flag
readop:     .ds 1           ; 1 if read operation
wrtype:     .ds 1           ; write operation type
dmaAddr:    .ds 2           ; last dma address
hstbuf:     .ds 512         ; host buffer

    .end